#if !defined(__UTILITY_H)
#define __UTILITY_H

class Font;
class Texture;

const int MAX_INTEGER = 0x7FFFFFFF;
const int MIN_INTEGER = 0x80000000;

#define OS_9X   0x0001
#define OS_NT   0x0002
#define OS_95   0x0004
#define OS_98   0x0008
#define OS_ME   0x0010
#define OS_NT4  0x0020
#define OS_2K   0x0040
#define OS_XP   0x0080
#define OS_2KXP 0x0100

enum
{
	MULTIBLT_STRETCH,
	MULTIBLT_TILE,
	MULTIBLT_TILEHORIZONTAL,
	MULTIBLT_TILEVERTICAL
};

struct NameValuePair
{
	const char *name;
	int value;
};

HWND GetLitestepDesktop();

bool IsAppWindow(const HWND &hWnd);
bool IsOS(const int &os);

void PrefixedLSSetVariable(string prefix, string var, string value);

string trim(const string &toTrim);
string lowerCase(const string &s);

void ModifyStyle(HWND hWnd, DWORD removeStyle, DWORD addStyle);
void ModifyStyleEx(HWND hWnd, DWORD removeStyle, DWORD addStyle);

bool GetxRCBool(const string &prefix, const string &baseName, bool defaultVal = false);

string GetxRCString(const string &prefix, const string &baseName, const string &defaultVal);

string GetxRCLine(const string &prefix, const string &baseName, const string &defaultVal);
StringList GetxRCLineList(const string &prefix);

StringList GetxRCCompleteList();

int GetxRCNamedValue(const string &prefix, const string &baseName, const NameValuePair *nameValuePairs, int defaultVal);
StringList GetxRCNameList(const string &prefix);

int GetxRCColor(const string &prefix, const string &baseName, int defaultVal);

int GetxRCInt(const string &prefix, const string &baseName, int defaultVal, int minVal = MIN_INTEGER, int maxVal = MAX_INTEGER);

int xParseCoord(const string &aString, int defaultVal, int maxVal, int lastVal = 0);
int GetxRCCoord(const string &prefix, const string &baseName, int defaultVal, int maxVal);

int xParseDimen(const string &aString, int defaultVal, int maxVal);
int GetxRCDimen(const string &prefix, const string &baseName, int defaultVal, int maxVal);

Font *GetRCFont(const string &prefix, Font *defaultVal = 0);

Texture *GetRCTexture(const string &prefix, const string &baseName, Texture *defaultVal = 0);

void PaintDesktopEx(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, int xSrc, int ySrc, BOOL updateCache);

void TileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp);
void HorizontalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp);
void VerticalTileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp);
void MultiBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, int cxLeft, int cyTop, int cxRight, int cyBottom, int mode, DWORD rasterOp);

LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue);

vector<string> split(string str, string sep);

BOOL GetClipboardText(HWND hwndOwner, LPTSTR pszBuffer, UINT cchBuffer);
BOOL SetClipboardText(HWND hwndOwner, LPCTSTR pszText);

string IntToString(int input);

#endif
